import os
import asyncio
import aiohttp
import itertools
import urllib.parse
import time
from dotenv import load_dotenv
from .models.option_models.universal_snapshot import UniversalOptionSnapshot
import pandas as pd
load_dotenv()
# Try to import a faster JSON library; fall back to built-in json if unavailable.
try:
    import orjson
    fast_json_loads = orjson.loads
except ImportError:
    import json
    fast_json_loads = json.loads

class PolygonOptions:
    def __init__(self):
        # Load environment variables from .env file
        load_dotenv()
        # Load API keys from environment variables (e.g., API_KEY_1, API_KEY_2, ...)
        self.api_keys = [
            os.getenv("YOUR_POLYGON_KEY"),
            os.getenv("YOUR_POLYGON_KEY2"),
            os.getenv("YOUR_POLYGON_KEY3"),
            os.getenv("YOUR_POLYGON_KEY4"),
            os.getenv("YOUR_POLYGON_KEY5"),
            os.getenv("YOUR_POLYGON_KEY6"),
            os.getenv("YOUR_POLYGON_KEY7"),
            os.getenv("YOUR_POLYGON_KEY8"),
            os.getenv("YOUR_POLYGON_KEY9"),
            os.getenv("YOUR_POLYGON_KEY10"),
            os.getenv("YOUR_POLYGON_KEY11"),
        ]
        # Remove any keys that weren't set
        self.api_keys = [key for key in self.api_keys if key]
        # Create a round-robin cycle for API keys
        self.api_key_cycle = itertools.cycle(self.api_keys)
        self.session = None

    def get_api_key(self) -> str:
        """Return the next API key from the cycle."""
        return next(self.api_key_cycle)

    def append_api_key(self, url: str, api_key: str) -> str:
        """
        Append an API key as a query parameter to the URL using URL encoding.
        If the URL already has query parameters, it will add to them.
        """
        parsed = urllib.parse.urlparse(url)
        query_params = urllib.parse.parse_qsl(parsed.query)
        query_params.append(('apiKey', api_key))
        new_query = urllib.parse.urlencode(query_params)
        new_url = urllib.parse.urlunparse((
            parsed.scheme,
            parsed.netloc,
            parsed.path,
            parsed.params,
            new_query,
            parsed.fragment
        ))
        return new_url

    async def get_http_session(self):
        """Return a shared aiohttp session with a high connection pool limit."""
        if self.session is None:
            connector = aiohttp.TCPConnector(limit=200)
            self.session = aiohttp.ClientSession(connector=connector)
        return self.session

    async def fetch_page(self, url: str, session: aiohttp.ClientSession) -> dict:
        """
        Fetch a single page from the API.
        Tries each API key until a successful (non-rate-limited) response is received.
        """
        for _ in range(len(self.api_keys)):  # Try each key once before failing
            api_key = self.get_api_key()
            url_with_key = self.append_api_key(url, api_key)
            try:
                async with session.get(url_with_key) as response:
                    if response.status == 429:
                        # Rate limited: wait a short time before trying a different key
                        await asyncio.sleep(0.25)
                        continue
                    if response.status != 200:
                        return {}
                    data = await response.json(loads=fast_json_loads)
                    return data
            except Exception as e:
                # In production, consider logging the exception.
                continue
        return {}

    async def fetch_all_pages(self, start_url: str) -> list:
        """
        Fetch all paginated results from the API starting at start_url.
        Uses a prefetch technique to overlap network latency.
        """
        results = []
        connector = aiohttp.TCPConnector(limit=200)
        async with aiohttp.ClientSession(connector=connector) as session:
            current_url = start_url
            # Prefetch the first page.
            prefetch_task = asyncio.create_task(self.fetch_page(current_url, session))
            while True:
                data = await prefetch_task
                if not data:
                    break
                if "results" in data:
                    results.extend(data["results"])
                # If there is no next_url, we are done.
                next_url = data.get("next_url")
                if not next_url:
                    break
                # Prefetch the next page while processing the current one.
                prefetch_task = asyncio.create_task(self.fetch_page(next_url, session))
                current_url = next_url
        return results

    async def get_option_chain_all(
        self,
        underlying_asset: str,
        strike_price: float = None,
        strike_price_lte: float = None,
        strike_price_gte: float = None,
        expiration_date: str = None,
        expiration_date_gte: str = None,
        expiration_date_lte: str = None,
        contract_type: str = None,
        order: str = None,
        limit: int = 250,
        sort: str = None
    ):
        """
        Retrieve all option contracts for the specified underlying asset across multiple pages.
        Optional parameters can be provided to filter the option chain.
        """
        if not underlying_asset:
            raise ValueError("Underlying asset ticker symbol must be provided.")

        # Remove "I:" prefix if present.
        if underlying_asset.startswith("I:"):
            underlying_asset = underlying_asset.replace("I:", "")

        # Build a dictionary of query parameters.
        params = {
            'strike_price': strike_price,
            'strike_price.lte': strike_price_lte,
            'strike_price.gte': strike_price_gte,
            'expiration_date': expiration_date,
            'expiration_date.gte': expiration_date_gte,
            'expiration_date.lte': expiration_date_lte,
            'contract_type': contract_type,
            'order': order,
            'limit': limit,
            'sort': sort
        }
        # Filter out parameters that are None.
        params = {k: v for k, v in params.items() if v is not None}

        # Construct the endpoint URL.
        endpoint = f"https://api.polygon.io/v3/snapshot/options/{underlying_asset}"
        if params:
            query_string = '&'.join(f"{key}={value}" for key, value in params.items())
            endpoint = f"{endpoint}?{query_string}"
        else:
            endpoint = f"{endpoint}?"

        # Append an API key to the starting URL.
        delimiter = '&' if '?' in endpoint else '?'
        endpoint = f"{endpoint}{delimiter}apiKey={self.get_api_key()}"

    
        results = await self.fetch_all_pages(endpoint)

        return UniversalOptionSnapshot(results)



    async def close(self):
        """Close the aiohttp session if it exists."""
        if self.session:
            await self.session.close()
